package graph;
import java.util.*;
import java.io.*;

public class Graph {
    private Map nodes = new TreeMap();
    private Map edges = new TreeMap();
    private Map impliedNodeProperties = new TreeMap();
    private ArrayList nodeTokens = new ArrayList();
    private ArrayList edgeTokens = new ArrayList();
    private String nodePrefix="", edgePrefix="";
    
    private void generateNodes(int amount) {
	for (int i=0; i < amount; i++) {
	    String label = nodePrefix+String.valueOf(i+1);
	    addNode(new Node(this,label));
	}	
    }
    final String delims = "\t\n\r\f\" ";
    private String flipDelims(String d) {
	if (d.equals(delims))
	    return "\"";
	else
	    return delims;
    }
    private ArrayList tokenize(String line) {
	ArrayList res = new ArrayList();
	String curDelims = delims;
	StringTokenizer t = new StringTokenizer(line,curDelims,true);
	while (t.hasMoreTokens()) {
	    String s = t.nextToken(curDelims);
	    if (curDelims.indexOf(s) == -1) {
		res.add(s);
	    }
	    else if (s.equals("\""))
		curDelims = flipDelims(curDelims);
	}
	return res;
    }
    private boolean parseFormats(String format) {
	ArrayList tokens = tokenize(format);
	if (tokens.size() != 6)
	    return false;
	StringTokenizer nodeFormat = new StringTokenizer((String)tokens.get(4));
	if (nodeFormat.countTokens() == 0)
	    generateNodes(Integer.parseInt((String)tokens.get(2)));
	else {
	    while (nodeFormat.hasMoreTokens()) 
		nodeTokens.add(nodeFormat.nextToken());
	}

	StringTokenizer edgeFormat = new StringTokenizer((String) tokens.get(5));
	while (edgeFormat.hasMoreTokens())
	    edgeTokens.add(edgeFormat.nextToken());
	return true;
    }
    private boolean parseProblem(String problem) {
	StringTokenizer t = new StringTokenizer(problem);
	if (t.countTokens() != 4)
	    return false;
	t.nextToken();
	String pid = t.nextToken();
	int numNodes = Integer.parseInt(t.nextToken());
	int numEdges = Integer.parseInt(t.nextToken());
	if (pid.equals("min") || pid.equals("max") || pid.equals("asn") || pid.equals("edge")) {
	    generateNodes(numNodes);
	}
	if (pid.equals("min")) {
	    nodeTokens.add("id");
	    nodeTokens.add("flow");

	    edgeTokens.add("source");
	    edgeTokens.add("dest");
	    edgeTokens.add("low");
	    edgeTokens.add("capacity");
	    edgeTokens.add("cost");
	} else if (pid.equals("max")) {
	    nodeTokens.add("id");
	    nodeTokens.add("endpoint");
	    
	    edgeTokens.add("source");
	    edgeTokens.add("dest");
	    edgeTokens.add("capacity");
	} else if (pid.equals("asn")) {
	    nodeTokens.add("id");
	    impliedNodeProperties.put("source","1");
	    
	    edgeTokens.add("source");
	    edgeTokens.add("dest");
	    edgeTokens.add("cost");
	} else if (pid.equals("edge")) {
	    edgeTokens.add("source");
	    edgeTokens.add("dest");
	}
	return true;
    }
    
    private int nodeIDCounter = 1;
    private boolean parseNode(String line) {
	ArrayList tokens = tokenize(line);
	if (tokens.size() != nodeTokens.size() +1)
	    return false;

	Node node = null;

	int idToken = nodeTokens.indexOf("id");
	if (idToken == -1) {
	    node = getNode(nodePrefix+String.valueOf(nodeIDCounter++));
	    if (node == null)
		return false;
	} else {
	    node = getNode(nodePrefix+(String)tokens.get(idToken+1));
	    if (node == null) {
		node = new Node(this,nodePrefix+(String)tokens.get(idToken+1));
		addNode(node);
	    }
	}
       
	for (int i=0; i < nodeTokens.size(); i++) {
	    if (i != idToken)
		node.set((String)nodeTokens.get(i),(String)tokens.get(i+1));
	}
	node.properties.putAll(impliedNodeProperties);
	return true;
    }
    
    private boolean parseEdge(String line) {
	ArrayList tokens = tokenize(line);
	if (tokens.size() != edgeTokens.size()+1)
	    return false;
	int srcToken = edgeTokens.indexOf("source");
	int dstToken = edgeTokens.indexOf("dest");
	if (dstToken == -1 || srcToken == -1)
	    return false;
	Node src = getNode(nodePrefix+(String)tokens.get(srcToken+1));
	Node dst = getNode(nodePrefix+(String)tokens.get(dstToken+1));
	if (dst == null || src == null)
	    return false;
	int idToken = edgeTokens.indexOf("id");
	String id = null;
	Edge edge = null;
	if (idToken == -1) {
	    edge = addEdge(src,dst);
	} else {
	    id = (String)tokens.get(idToken+1);
	    edge = addEdge(edgePrefix + id,src,dst);
	}
	
	for (int i=0; i < edgeTokens.size(); i++) {
	    if (i != idToken || i != srcToken || i != dstToken)
		edge.set((String)edgeTokens.get(i),(String)tokens.get(i+1));
		
	}
	return true;
    }
    /**
     *Initializes a new Graph by parsing the graph-file from the argument stream.
     *Prefixes each node and edge by strings given as arguments. This is useful
     *when input nodes or edges do not have labels and are automatically assigned
     *a numeric label.
     *@param stream an InputStream to be parsed
     *@param nodePrefix a String to be used as a prefix for the node labels
     *@param edgePrefix a String to be used as a prefix for the edge labels
     */
    public Graph(InputStream stream,String nodePrefix, String edgePrefix) {
	this.nodePrefix = nodePrefix;
	this.edgePrefix = edgePrefix;
	load(stream);
    }
    /**
     *Initializes a new Graph by parsing the graph-file from the argument stream
     *@param stream InputStream to be parsed
     */
    public Graph(InputStream stream) {
	load(stream);
    }

    private void load(InputStream file) {
	try {
	    BufferedReader reader = 
		new BufferedReader(new InputStreamReader(file));
	    while(true) {
		String line = reader.readLine();
		if (line == null)
		    break;
		StringTokenizer t = new StringTokenizer(line);
		if (!t.hasMoreTokens())
		    continue;
		String id = t.nextToken();
		if (id.equals("c")) {
		    if (t.hasMoreTokens() && t.nextToken().equals("f")) {
			parseFormats(line);
		    } else 
			continue;
		}
		if (id.equals("p")) {
		    parseProblem(line);
		    continue;
		}
		if (id.equals("n") || id.equals("v")) {
		    parseNode(line);
		    continue;
		}
		if (id.equals("e") || id.equals("a")) {
		    parseEdge(line);
		    continue;
		}
	    }
	} catch (IOException e) {
	    System.out.println(e);
	}
    }
    /**
     *Returns the Node in the Graph with label given as argument
     *@param label label of the Node
     *@return a Node or null if there is no Node with given label
     */
    public Node getNode(String label) {
	return (Node)nodes.get(label);
    }

    /**
     *Returns the Edge in the Graph with label given as argument
     *@param label label of the Edge
     *@return an Edge or null if there is no Edge with given label
     */
    public Edge getEdge(String label) {
	return (Edge)nodes.get(label);
    }

    /**
     *Returns an ArrayList containing all the nodes in the graph
     *@return an ArrayList of Nodes.
     */
    public ArrayList getNodes() {
	ArrayList res = new ArrayList();
	Collection c = nodes.values();
	Iterator i = c.iterator();
	while (i.hasNext())
	    res.add(i.next());
	return res;
    }

    /**
     *Returns an ArrayList containing all the edges in the graph
     *@return an ArrayList of Edges.
     */
    public ArrayList getEdges() {
	ArrayList res = new ArrayList();
	Iterator i = edges.values().iterator();
	while (i.hasNext())
	    res.add(i.next());
	return res;
    }

    /**
     *Adds a Node to the graph
     *@param node a node to be added.
     */
    public void addNode(Node node) {
	nodes.put(node.label(),node);
    }

    private int edgeIDCounter = 1;
    /**
     *Adds an Edge to the graph. The new edge goes from source to dest node.
     *A label is generated for the edge.
     *@param source the source node of the edge
     *@param dest the destination node of the edge
     *@return the newly created Edge.
     */
    public Edge addEdge(Node source, Node dest) {
	while(edges.get(edgePrefix+String.valueOf(edgeIDCounter)) != null) {
	    edgeIDCounter++;
	}
	return addEdge(edgePrefix+String.valueOf(edgeIDCounter),source,dest);
    }
    /**
     *Adds an Edge to the graph. The new edge goes from source to dest node.
     *@param label the label for the new edge
     *@param source the source node of the edge
     *@param dest the destination node of the edge
     *@return the newly created Edge or null if edge with given label exists.
     */
    public Edge addEdge(String label, Node source, Node dest) {
	if (edges.get(label) != null)
	    return null;
	Edge e = new Edge(label, source,dest);
	edges.put(e.label(),e);
	source.addOutEdge(e);
	dest.addInEdge(e);
	return e;
    }
    /**
     *Returns an ArrayList containing all the nodes in the graph that have
     *a given property with a given value
     *@param property a String containing the requested property
     *@param value a String containing the requested value
     *@return an ArrayList of Nodes.
     */
    public ArrayList getNodes(String property, String value) {
	ArrayList res = new ArrayList();
	Iterator i = nodes.values().iterator();
	while (i.hasNext()) {
	    Node n = (Node)i.next();
	    String v = n.get(property);
	    if (v != null && v.equals(value))
		res.add(n);
	}
	return res;
    }

    /**
     *Returns an ArrayList containing all the edges in the graph that have
     *a given property with a given value
     *@param property a String containing the requested property
     *@param value a String containing the requested value
     *@return an ArrayList of Edges.
     */
    public ArrayList getEdges(String property, String value) {
	ArrayList res = new ArrayList();
	Iterator i = edges.values().iterator();
	while (i.hasNext()) {
	    Edge e = (Edge)i.next();
	    String v = e.get(property);
	    if (v != null && v.equals(value))
		res.add(e);
	}
	return res;
    }
}
